(function($console) {
	'use strict';

	window.ipConfigure = window.ipConfigure || {};
	ipConfigure.components = ipConfigure.components || {};

	/**
	 * Compatability-mode video class, injects an IMG into the element indicated by the id parameter.
	 * Exposes events and allows control of the "video".
	 *
	 * @param {string} id
	 * @param {object} options backgroundColor (html color), debug (bool), requestBaseUrl (callback)
	 *
	 * @namespace ipConfigure.components.video
	 */
	ipConfigure.components.videoCompatibility = function(id, options) {
		var that = this,
			_pluginContainer = document.querySelector('#' + id),
			_imgContainer,
			_img,                           // the img element
			_videoMask,                     // semi-transparent mask that sits on the top of the video, dimming it
			_statusOverlay,                 // spinning asterisk indicating state (connecting or buffering)
			_spinInner,
			_sized = false,                 // determines whether or not the stream size has been detected
			_sizingInfo = {},
			_clockTime,
			_currentTime,
			_isBuffering = false,
			_srcTimer,
			_resizeTimer,
			_renegotiateSizerTimer,
			_cachedBaseUrl = '',
			_cachedFps = 0,
			_defaultOptions = {
				backgroundColor: '#EEEEEE',
				debug: false,
				getTime: function() {
					_error('getTime event not subscribed to');
					return 0;
				},
				requestBaseUrl: function() {
					_error('requestBaseUrl event not subscribed to');
					return 'http://notfound';
				},
				onAudioKnown: function(hasAudio) {
					_log('onAudioKnown, hasAudio: ', hasAudio);
				},
				onEndReached: function() {
				},
				onErrorEncountered: function() {
				},
				onLoading: function() {
				},
				onPlaying: function() {
				},
				onNotification: function(s) {
					_log('onNotification event not subscribed to: ', s);
				},
				onPlayFromTime: function() {
				},
				onPlayLive: function() {
				},
				stream: undefined
			},
			_options = ipConfigure.core.mergeOptions(_defaultOptions, options);

		// private functions

		function _error() {
			$console.error.apply($console, arguments);
		}

		function _log() {
			if (_options.debug === true) {
				$console.log.apply($console, arguments);
			}
		}

		function _createComponent() {
			var html = ' ' +
				'<div class="plugin" style="width: 100%; height: 100%; background-color: ' + _options.backgroundColor + '"><img></div>' +
				'<div class="mask"></div>' +
				'<div class="status">' +
				'   <span class="spin-outer">' +
				'       <i class="fa fa-asterisk fa-2x spin-inner"></i>' +
				'   </span>' +
				'</div>';

			_pluginContainer.className += ' ipc-video';
			_pluginContainer.innerHTML = html;

			_imgContainer = _pluginContainer.querySelector('.plugin');
			_img = _pluginContainer.querySelector('img');
			_statusOverlay = _pluginContainer.querySelector('.status');
			_spinInner = _pluginContainer.querySelector('.status .spin-inner');
			_videoMask = _pluginContainer.querySelector('.mask');
		}

		function _setSize(skipRenegotiation) {
			var containerAspectRatio,
				favor;

			// cancel previous renegotiate attempt
			if (_renegotiateSizerTimer) {
				clearTimeout(_renegotiateSizerTimer);
			}

			if (_sized === true) {
				containerAspectRatio = _imgContainer.clientWidth / _imgContainer.clientHeight;

				favor = _sizingInfo.aspectRatio > containerAspectRatio ? 'width' : 'height';

				// figure out if we need to favor width or height
				_img.style[favor === 'width' ? 'height' : 'width'] = '';
				_img.style[favor] = '100%';

				// renegotiate resize size 5 seconds after this function has run
				if (skipRenegotiation !== true && that.isPlaying === true) {
					_renegotiateSizerTimer = setTimeout(function() {
						_negotiateBaseUrl(that.isLive === true ? 0 : _options.getTime(), 'resize');
					}, 2000);
				}
			}
		}

		function _onImgLoad() {
			var minWait = 33,
				wait = minWait;

			if (_isBuffering === true) {
				_isBuffering = false;

				_toggleDither(false);
				_toggleLoadingState(0);
			}

			if (_sized === false) {
				_options.stream.aspectRatio = [_img.clientWidth, _img.clientHeight];
				_sized = true;
				_sizingInfo.aspectRatio = _img.naturalWidth / _img.naturalHeight;
				_setSize(true);
			}

			_hideNotification();
			_img.style.display = '';
			_imgContainer.style.opacity = 1;

			// delay based on FPS, sometimes 0 is provided... 1000 / 0 is Infinity... so don't let that happen
			if (_cachedFps === 0) {
				wait = 1000;
			}
			else if (_cachedFps > 0) {
				wait = Math.max(minWait, (1000 / _cachedFps));
			}

			_setSrcDelayed(wait);
		}

		function _onImgLoadError() {
			_log('MediaPlayerEncounteredError');
			_showNotification('Unable to display video at this time');
			_img.style.display = 'none';

			_toggleLoadingState(0);

			_options.onErrorEncountered();

			// no longer needed, client should renegotiate
			//_setSrcDelayed(5000);
		}

		/**
		 * @param {number} state 0 = done, 1 = loading, 2 = buffering
		 */
		function _toggleLoadingState(state) {
			switch (state) {
				case 0:
					_statusOverlay.style.display = '';
					break;
				case 1:
					_spinInner.className = _spinInner.className.replace('spin-inner-stop', '').trim();
					_statusOverlay.style.display = 'block';
					break;
				case 2:
					if (_spinInner.className.indexOf('spin-inner-stop') === -1) {
						_spinInner.className += ' spin-inner-stop';
					}
					_statusOverlay.style.display = 'block';
					break;
			}
		}

		function _toggleDither(dither) {
			_videoMask.style.opacity = ((dither === true) ? 0.5 : 0);
		}

		function _showNotification(text) {
			_toggleDither(true);
			_options.onNotification(text);
		}

		function _hideNotification() {
			_toggleDither(false);
			_options.onNotification('');
		}

		function _setSrc() {
			if (_cachedBaseUrl !== '') {
				_img.src = _cachedBaseUrl + '?' + Date.now();
			}
		}

		function _setSrcDelayed(delayBetweenFrames) {
			_srcTimer = setTimeout(function() {
				_currentTime += Date.now() - _clockTime;
				_clockTime = Date.now();

				if (that.isPlaying === true) {
					_setSrc();
				}
			}, delayBetweenFrames);
		}

		function _stop() {
			_cachedBaseUrl = '';

			clearTimeout(_srcTimer);
		}

		function _negotiateBaseUrl(startTime, via) {
			_log('NEGOTIATE NOW', startTime);

			_options.requestBaseUrl({
				start: startTime,
				width: _imgContainer.parentElement.parentElement.clientWidth,
				height: _imgContainer.parentElement.parentElement.clientHeight,
				via: via
			}, function(baseUrl, fpsOrReason) {
				if (that.isPlaying === true) {
					if (baseUrl === false) {
						_cachedBaseUrl = '';

						// TODO trash
						if (fpsOrReason && fpsOrReason.indexOf('has been reached') > -1) {
							_showNotification(fpsOrReason);
						}
						else {
							_showNotification('Unable to display video at this time');
						}
						_img.style.display = 'none';

						_toggleLoadingState(0);

						_options.onErrorEncountered();
					}
					else if (baseUrl === '') {
						_isBuffering = true;
						_toggleDither(true);
						_toggleLoadingState(1);

						_cachedBaseUrl = baseUrl;
						_cachedFps = fpsOrReason;
					}
					else {
						_cachedBaseUrl = baseUrl;
						_cachedFps = fpsOrReason;
						_setSrc();
					}
				}
			});
		}

		function _init() {
			_log('video options', _options);

			// create the img tag and overlays
			_createComponent();

			// create subscriptions to necessary events
			_img.addEventListener('load', _onImgLoad);
			_img.addEventListener('error', _onImgLoadError);

			// monitor for changes in size, so we can request an appropriately sized image from the server
			_resizeTimer = setInterval(function tick() {
				var firstTime = angular.isUndefined(_sizingInfo.containerWidth),
					width = _imgContainer.parentElement.parentElement.clientWidth,
					height = _imgContainer.parentElement.parentElement.clientHeight;

				if (_sizingInfo.containerWidth !== width || _sizingInfo.containerHeight !== height) {
					_sizingInfo.containerWidth = width;
					_sizingInfo.containerHeight = height;

					_log('SET SIZE FIRST TIME', firstTime);
					_setSize(firstTime);
				}
			}, 333);
		}

		// public properties

		that.isLive = false;

		that.isPlaying = false;

		that.options = _options;

		// public methods

		that.getVolume = function() {
			return 0;
		};

		that.setVolume = function() {
			// placebo
		};

		that.getCurrentTime = function() {
			return _currentTime;
		};

		that.getVideo = function() {
			return _img;
		};

		/**
		 * @param {number} startTime number of milliseconds since 1970
		 * @param {boolean} autoPause automatically pauses the video when it begins playing when true
		 */
		that.playFromTime = function(startTime, autoPause) {
			if (typeof _options.stream === 'undefined') {
				_error('videoCompatibility.options.stream not set');
			}
			else {
				// reset locals in case the video dimensions have changed
				_sized = false;

				// set the loading state and stop whatever is playing
				_clockTime = Date.now();
				_currentTime = startTime;
				_isBuffering = true;
				_toggleDither(true);
				_toggleLoadingState(1);
				_stop();

				// set necessary properties
				that.isLive = false;
				that.isPlaying = autoPause ? false : true;

				// let em know
				_options.onPlayFromTime();

				// start pulling frames
				_negotiateBaseUrl(startTime, 'play');
			}
		};

		that.playLive = function() {
			if (typeof _options.stream === 'undefined') {
				_error('videoCompatibility.options.stream not set');
			}
			else {
				// reset locals in case the video dimensions have changed
				_sized = false;

				// set necessary properties
				that.isLive = true;
				that.isPlaying = true;

				// set the loading state and stop whatever is playing
				_clockTime = Date.now();
				_currentTime = _clockTime;
				_isBuffering = true;
				_toggleDither(true);
				_toggleLoadingState(1);
				_stop();

				// let em know
				_options.onPlayLive();

				// start pulling frames
				_negotiateBaseUrl(0, 'play');
			}
		};

		that.stop = function() {
			// clean up public properties
			that.isLive = false;
			that.isPlaying = false;

			// clean up the video
			_hideNotification();

			_stop();
		};

		that.destroy = function() {
			_log('video destroy');

			that.isPlaying = false;

			if (_srcTimer) {
				clearTimeout(_srcTimer);
			}

			if (_renegotiateSizerTimer) {
				clearTimeout(_renegotiateSizerTimer);
			}

			clearInterval(_resizeTimer);

			_img.removeEventListener('load', _onImgLoad, false);
			_img.removeEventListener('error', _onImgLoadError, false);
		};

		// initialize

		_init();
	};
}(console));
